# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011, Ansgar Burchardt <ansgar@debian.org>
#
# Released under the same terms as the original software, see below.
#
# Based on quoted_regex_parse from uscan which has the following copyright
# notice:
#
# Originally written by Christoph Lameter <clameter@debian.org> (I believe)
#
# Modified by Julian Gilbey <jdg@debian.org>
#
# HTTP support added by Piotr Roszatycki <dexter@debian.org>
#
# Rewritten in Perl, Copyright 2002-2006, Julian Gilbey
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from pet.exceptions import *

import re

_re_op = re.compile(r'^(s|tr|y)(.*)$')
_markers = { '{': '}', '(': ')', '[': ']' }

_pattern_rules = [
  # POSIX character classes
  (re.compile(r'\[:alpha:\]'), r'A-Za-z'),
  (re.compile(r'\[:alnum:\]'), r'A-Za-z0-9'),
    # missing: [:ascii:]
  (re.compile(r'\[:blank:\]'), r' \t'),
    # missing: [:cntrl:]
  (re.compile(r'\[:digit:\]'), r'\d'),
    # missing: [:graph:]
  (re.compile(r'\[:lower:\]'), r'a-z'),
    # missing: [:print:], [:punct:]
  (re.compile(r'\[:space:\]'), r'\s'), # missing: + vertical tab
  (re.compile(r'\[:upper:\]'), r'A-Z'),
  (re.compile(r'\[:word:\]'), r'\w'),
  (re.compile(r'\[:xdigit:\]'), r'0-9a-fA-F'),
]

_replacement_rules = [
  # backrefs
  (re.compile(r'\$&'), r'\\g<0>'),
  (re.compile(r'\$\{?(\d)\}?'), r'\\g<\1>'),
]

def compile(pattern):
  for regex, sub in _pattern_rules:
    pattern = regex.sub(sub, pattern)
  return re.compile(pattern)

def apply_perlre(regexp, string):
  regexp = regexp.strip()
  if regexp == "":
    return string

  match_op = _re_op.match(regexp)
  if not match_op:
    raise RegexpError("Unknown operator in regular expression '{0}'.".format(regexp))
  op = match_op.group(1)

  if op != 's':
    raise NotImplemented("Operator '{0}' not implemented.".format(op))

  arguments = match_op.group(2)
  marker = arguments[0]
  end_marker = _markers.get(marker, marker)
  last_was_escape = False

  stage = 0 # 0=pattern, (1=replacement-marker), 2=replacement, 3=flags
  pattern = replacement = flags = ""
  for char in arguments[1:]:
    if stage == 1:
      end_marker = _markers.get(char, char)
      stage = 2
      continue
    if last_was_escape:
      last_was_escape = False
      if stage == 0:
        pattern += char
      elif stage == 1:
        raise RegexpError("Invalid regular expression.")
      elif stage == 2:
        replacement += char
      else:
        flags += char
    elif char == end_marker:
      if stage == 0:
        if marker != end_marker:
          stage = 1
        else:
          stage = 2
      else:
        stage = 3
    else:
      if char == "\\":
        last_was_escape = True
      if stage == 0:
        pattern += char
      elif stage == 2:
        replacement += char
      else:
        flags += char

  if stage != 3:
    raise RegexpError("Invalid regular expression.")

  count = 1
  py_flags = 0
  for flag in flags:
    if flag == 'i':
      py_flags |= re.I
    elif flag == 'g':
      count = 0
    else:
      raise RegexpError("Unknown flag '{0}' used in regular expression.".format(flags))

  for regex, sub in _pattern_rules:
    pattern = regex.sub(sub, pattern)
  for regex, sub in _replacement_rules:
    replacement = regex.sub(sub, replacement)

  # TODO: flags is only in Python 2.7
  try:
    return re.sub(pattern, replacement, string, count=count)
  except:
    return string
  #return re.sub(pattern, replacement, string, count=count, flags=py_flags)



# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011, Ansgar Burchardt <ansgar@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import debianbts
from pet.exceptions import *

class _BugReport(object):
  def update_bug(self, bug):
    from pet.models import Session, BugSource
    session = Session.object_session(bug)

    # TODO: do this in a transaction; move creating a bug here (if None is passed in)
    try:
      assert bug.bug_number == self.bug_number
      bug.severity = self.severity
      bug.tags = self.tags
      bug.subject = self.subject
      bug.submitter = self.submitter
      bug.merged_with = self.merged_with
      bug.created = self.created
      bug.last_modified = self.last_modified
      bug.done = self.done
      bug.forwarded = self.forwarded
      bug.blocks = self.blocks
      bug.blocked_by = self.blocked_by
      bug.owner = self.owner
      bug.affects = bug

      fixed = self.fixed_versions
      found = self.found_versions
      session.query(BugSource).filter_by(bug=bug).delete()
      bug_sources = list()
      for s in self.sources:
        bs = BugSource(bug=bug, source=s, fixed_versions=fixed.get(s, []), found_versions=found.get(s, []))
        bug_sources.append(bs)
      bug.bug_sources = bug_sources

    except:
      raise

class _DebianBugReport(_BugReport):
  def __init__(self, bugreport, binary_source_map, ignore_unknown_binaries=False):
    self._bugreport = bugreport
    self._binary_source_map = binary_source_map
    self._ignore_unknown_binaries = ignore_unknown_binaries

  bug_number = property(lambda self: self._bugreport.bug_num)
  severity = property(lambda self: self._bugreport.severity)
  tags = property(lambda self: self._bugreport.tags)
  subject = property(lambda self: self._bugreport.subject)
  submitter = property(lambda self: self._bugreport.originator)
  merged_with = property(lambda self: self._bugreport.mergedwith)

  @property
  def sources(self):
    return [ s.strip() for s in self._bugreport.source.split(",") ]

  created = property(lambda self: self._bugreport.date)
  last_modified = property(lambda self: self._bugreport.log_modified)
  done = property(lambda self: self._bugreport.done)

  def _split_versions(self, version_strings):
    """
    given a list of strings "binary/version" or "version", this function
    returns a dict mapping source package names to listed versions.
    For strings of the format "version", it is assumed that all source
    packages for this bug report are meant.
    """
    versions = {}
    for version in version_strings:
      parts = version.split("/")
      if len(parts) == 1:
        sources = self.sources
        v = parts[0]
      elif len(parts) == 2:
        v = parts[1]
        try:
          sources = self._binary_source_map[parts[0]]
        except KeyError:
          if parts[0] in self.sources:
            # assume binary package name == source package name
            sources = [ parts[0] ]
          elif self._ignore_unknown_binaries:
            continue
          else:
            raise BinaryNotKnown("source for binary {0} is unknown".format(parts[0]))
      else:
        raise BugTrackerException("cannot parse version '{0}'".format(version))
      for s in sources:
        sv = versions.setdefault(s, set())
        sv.add(v)
    return versions
  @property
  def fixed_versions(self):
    return self._split_versions(self._bugreport.fixed_versions)
  @property
  def found_versions(self):
    return self._split_versions(self._bugreport.found_versions)

  forwarded = property(lambda self: self._bugreport.forwarded)
  blocks = property(lambda self: self._bugreport.blocks)
  blocked_by = property(lambda self: self._bugreport.blockedby)
  owner = property(lambda self: self._bugreport.owner)
  affects = property(lambda self: self._bugreport.affects)
  summary = property(lambda self: self._bugreport.summary)

class DebianBugTracker(object):
  def __init__(self, binary_source_map, ignore_unknown_binaries=False):
    self.binary_source_map = binary_source_map
    self.ignore_unknown_binaries = ignore_unknown_binaries
  def search(self, sources, bug_numbers=None):
    if bug_numbers is None:
      bug_numbers = set()
    else:
      bug_numbers = set(bug_numbers)
    bug_numbers.update(debianbts.get_bugs('src', sources))

    # process in batches. stolen from python-debianbts.
    BATCH_SIZE = 500
    result = []
    for i in range(0, len(bug_numbers), BATCH_SIZE):
        slice = list(bug_numbers)[i:i + BATCH_SIZE]
        result += debianbts.get_status(slice)

    return [ _DebianBugReport(b, self.binary_source_map, self.ignore_unknown_binaries) for b in result ]



# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011-2012, Ansgar Burchardt <ansgar@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from pet.exceptions import *

import json
import StringIO
import subprocess
import svn.client
import svn.core
import svn.ra
import time
import urllib2
import urllib

_vcs_backends = {}
def _vcs_backend(name):
  def helper(cls):
    global _vcs_backends
    _vcs_backends[name] = cls
    return cls
  return helper

def vcs_backend(repository):
  return _vcs_backends[repository.type](repository)

class VCS(object):
  pass

class _SubversionCallbacks(svn.ra.Callbacks):
  def __init__(self):
    self.auth_baton = svn.core.svn_auth_open([
      svn.client.get_simple_provider(),
      svn.client.get_username_provider(),
      ])

@_vcs_backend("svn")
class Subversion(VCS):
  def __init__(self, repository):
    self.root = repository.root
    self.web_root = repository.web_root
    self.callbacks = _SubversionCallbacks()
    self.ra = svn.ra.svn_ra_open2(self.root, self.callbacks, None)
    self.rev = svn.ra.svn_ra_get_latest_revnum(self.ra)
    self._cache = dict()
  def link(self, package, filename, directory=False, branch=None, tag=None, named_tree=None):
    assert not (named_tree and (branch or tag)), "cannot give both named_tree and branch or tag"
    if named_tree is not None:
      if named_tree.type == 'branch':
        branch = named_tree.name
      elif named_tree.type == 'tag':
        tag = named_tree.name
      else:
        raise ValueError('NamedTree is neither branch nor tag')
    assert not (branch and tag), "cannot give both branch and tag"
    if branch:
      prefix = "branches/{0}/{1}".format(branch, package)
    elif tag:
      prefix = "tags/{0}/{1}".format(package, tag)
    else:
      prefix = "trunk/{0}".format(package)

    if directory:
      extra = ""
    else:
      extra = "?view=markup"

    return "{0}/{1}/{2}{3}".format(self.web_root, prefix, filename, extra)
  def file(self, package, filename, branch=None, tag=None):
    """
    returns file contents
    """
    assert not (branch and tag), "cannot give both branch and tag"
    if branch:
      path = "branches/{1}/{0}/{2}".format(package, branch, filename)
    elif tag:
      path = "tags/{0}/{1}/{2}".format(package, tag, filename)
    else:
      path = "trunk/{0}/{1}".format(package, filename)
    try:
      stream = StringIO.StringIO()
      svn.ra.svn_ra_get_file(self.ra, path, self.rev, stream)
    except svn.core.SubversionException as e:
      if e.apr_err == 160013:
        raise FileNotFound(e.message)
      raise VCSException(e)
    return stream.getvalue()
  def _list(self, path):
    """
    retrieve list of (name, commit) of subdirectories in path
    """
    if path not in self._cache:
      entries = svn.ra.svn_ra_get_dir2(self.ra, path, self.rev, svn.core.SVN_DIRENT_KIND|svn.core.SVN_DIRENT_CREATED_REV)
      cache = dict()
      for name, dirent in entries[0].items():
        if dirent.kind != svn.core.svn_node_dir: continue
        cache[name] = dirent.created_rev
      self._cache[path] = cache

    return self._cache[path]
  @property
  def packages(self):
    """
    returns a list of known packages in the repository.
    """
    return self._list("trunk").keys()
  def branches(self, package):
    root = self._list("")
    trunk = self._list("trunk")
    branches = { None: trunk[package] }
    all_branches = self._list("branches")
    for name, commit in all_branches.items():
      branch = self._list("branches/{0}".format(name))
      if package in branch:
        branches[name] = branch[package]
    return branches
  def tags(self, package):
    return self._list("tags/{0}".format(package))
  def changed_named_trees(self, session, named_trees_by_package):
    # XXX: This function is ugly.
    from pet.models import NamedTree
    # We assume that all named_trees for a package were updated at the same time.
    # This means we can skip looking into a directory, if we already have a
    # named_tree with a higher commit_id.
    changed = {}
    trunk = self._list("trunk")
    tags = self._list("tags")
    branches = self._list("branches")

    for package, nts in named_trees_by_package.items():
      def add_changed(named_tree):
        changed.setdefault(package, []).append(named_tree)

      package_name = package.name
      known_tags = {}
      known_branches = {}
      known_trunk = None

      for nt in nts:
        if nt.type == 'tag':
          known_tags[nt.name] = nt
        elif nt.type == 'branch' and nt.name is not None:
          known_branches[nt.name] = nt
        elif nt.type == 'branch' and nt.name is None:
          known_trunk = nt
        else:
          raise Exception("unknown named_tree type (type={0}, name={1})".format(nt.type, nt.name))

      if package_name not in tags:
        for nt in known_tags.itervalues():
          session.delete(nt)
      else:
        if len(known_tags) > 0:
          highest_tag = int(max(known_tags.itervalues(), key=lambda nt: int(nt.commit_id)).commit_id)
        else:
          highest_tag = -1
        if tags[package_name] > highest_tag:
          existing_tags = self._list("tags/{0}".format(package_name))
          for name, nt in known_tags.iteritems():
            commit_id = existing_tags.get(name, None)
            if commit_id is None:
              session.delete(nt)
            elif commit_id > int(nt.commit_id):
              nt.commit_id = str(commit_id)
              add_changed(nt)

          for name, commit_id in existing_tags.iteritems():
            if name not in known_tags:
              nt = NamedTree(package=package, type='tag', name=name, commit_id=str(commit_id))
              add_changed(nt)
              session.add(nt)

      for branch_name in branches.iterkeys():
        branch = self._list("branches/{0}".format(branch_name))
        commit_id = branch.get(package_name, None)
        if commit_id is not None:
          nt = known_branches.get(branch_name, None)
          if nt is None:
            nt = NamedTree(package=package, type='branch', name=branch_name, commit_id=str(commit_id))
            add_changed(nt)
            session.add(nt)
          elif commit_id > int(nt.commit_id):
            nt.commit_id = str(commit_id)
            add_changed(nt)
        else:
          nt = known_branches.get(branch_name, None)
          if nt is not None:
            session.delete(nt)

      if known_trunk is None:
        if package_name in trunk:
          nt = NamedTree(package=package, type='branch', name=None, commit_id=str(trunk[package_name]))
          add_changed(nt)
          session.add(nt)
      else:
        commit_id = trunk.get(package_name, None)
        if commit_id is None:
          session.delete(known_trunk)
        elif commit_id > int(known_trunk.commit_id):
          known_trunk.commit_id = str(commit_id)
          add_changed(known_trunk)

    return changed

@_vcs_backend("git")
class Git(VCS):
  def __init__(self, repository):
    self.root = repository.root
    self.web_root = repository.web_root
    self._summary_cache = None
  def link(self, package, filename=None, directory=False, branch=None, tag=None, named_tree=None):
    assert not (named_tree and (branch or tag)), "cannot give both named_tree and branch or tag"
    if named_tree is not None:
      if named_tree.type == 'branch':
        branch = named_tree.name
      elif named_tree.type == 'tag':
        tag = named_tree.name
      else:
        raise ValueError('NamedTree is neither branch nor tag')
    assert not (branch and tag), "cannot give both branch and tag"
    if filename is not None and not directory:
      url = "/{0}.git;a=blob;f={1}".format(urllib.quote(package), urllib.quote(filename))
    elif filename is not None and directory:
      url = "/{0}.git;a=tree;f={1}".format(urllib.quote(package), urllib.quote(filename))
    else:
      url = ";a=shortlog"

    if branch:
      url += ";hb=refs/heads/{0}".format(urllib.quote(branch))
    elif tag:
      url += ";hb=refs/tags/{0}".format(urllib.quote(tag))

    return self.web_root + url
  def file(self, package, filename, branch=None, tag=None):
    """
    returns file contents
    """
    assert not (branch and tag), "cannot give both branch and tag"
    if branch is not None:
      extra = ';hb=refs/heads/{0}'.format(urllib.quote(branch))
    elif tag is not None:
      extra = ';hb=refs/tags/{0}'.format(urllib.quote(tag))
    else:
      extra = ""
    url = "{0}/{1}.git;a=blob_plain;f={2}{3}".format(self.web_root, urllib.quote(package), urllib.quote(filename), extra)

    try:
      f = urllib2.urlopen(url)
      contents = f.read()
      f.close()
    except urllib2.HTTPError as e:
      if e.code == 404:
        return None
      raise
    time.sleep(3)
    return contents
  @property
  def _summary(self):
    if self._summary_cache is None:
      f = urllib2.urlopen(self.root)
      contents = f.read()
      f.close()
      self._summary_cache = json.loads(contents)
    return self._summary_cache
  @property
  def packages(self):
    return self._summary.keys()
  def branches(self, package):
    branches = self._summary[package]['branches']
    branches[None] = self._summary[package]['trunk']
    return branches
  def tags(self, package):
    return self._summary[package]['tags']
  def changed_named_trees(self, session, named_trees_by_package):
    # XXX: This function is ugly.
    from pet.models import NamedTree
    s = self._summary

    changed = {}

    for package, nts in named_trees_by_package.iteritems():
      def add_changed(named_tree):
        changed.setdefault(package, []).append(named_tree)
      ps = s.get(package.name)
      if ps is None:
        for nt in nts:
          session.delete(nt)
      else:
        nts_by_type = dict(branch={}, tag={})
        trunk = None
        for nt in nts:
          if nt.type == 'branch' and nt.name is None:
            trunk = nt
          else:
            nts_by_type[nt.type][nt.name] = nt
        # add an alias for now...
        ps['tag'] = ps['tags']
        ps['branch'] = ps['branches']
        for type in ('branch', 'tag'):
          nts = nts_by_type[type]
          known = ps[type]
          for name, commit_id in known.iteritems():
            nt = nts.get(name, None)
            if nt is None:
              nt = NamedTree(package=package, type=type, name=name, commit_id=commit_id)
              add_changed(nt)
              session.add(nt)
            elif nt.commit_id != commit_id:
              nt.commit_id = commit_id
              add_changed(nt)
          for name, nt in nts.iteritems():
            if name not in known:
              session.delete(nt)
        if trunk is None:
          trunk = NamedTree(package=package, type='branch', name=None, commit_id=ps['trunk'])
          add_changed(trunk)
          session.add(trunk)
        elif trunk.commit_id != ps['trunk']:
          trunk.commit_id = ps['trunk']
          add_changed(trunk)
    return changed

@_vcs_backend("git-local")
class GitLocal(Git):
  def __init__(self, repository):
    super(GitLocal, self).__init__(repository)
  def file(self, package, filename, branch=None, tag=None):
    assert not (branch and tag), "cannot give both branch and tag"
    tree = branch or tag or 'HEAD'
    cmd = ['git', 'cat-file', 'blob', '{0}:{1}'.format(tree, filename)]
    cwd = '{0}.d/{1}.git'.format(self.root, package)
    p = subprocess.Popen(cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    if p.returncode != 0:
      return None
    return stdout
  @property
  def _summary(self):
    if self._summary_cache is None:
      with open(self.root, 'r') as fh:
        contents = fh.read()
        self._summary_cache = json.loads(contents)
    return self._summary_cache

@_vcs_backend("git-ssh")
class GitSsh(Git):
  def __init__(self, repository):
    super(GitSsh, self).__init__(repository)
  def file(self, package, filename, branch=None, tag=None):
    assert not (branch and tag), "cannot give both branch and tag"
    tree = branch or tag or 'HEAD'
    cwd = '{0}/{1}.git'.format(self.root, package)
    cmd = ['ssh', 'pet-cat-file', 'cd', cwd, ';', 'git', 'cat-file', 'blob', '{0}:{1}'.format(tree, filename)]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = p.communicate()
    if p.returncode != 0:
      return None
    return stdout
  @property
  def _summary(self):
    if self._summary_cache is None:
      cmd = ['ssh', 'pet-cat-file', 'cd', self.root, ';', 'pet-summary']
      p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      (stdout, stderr) = p.communicate()
      if p.returncode != 0:
        raise Exception("Getting summary for {0} failed: {1}".format(self.root, stderr))
      self._summary_cache = json.loads(stdout)
    return self._summary_cache



# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011, Ansgar Burchardt <ansgar@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sqlalchemy
# from serve import pet_serve
from .. import pet_update

def engine(no_cert):
    if(no_cert):
        return sqlalchemy.create_engine('postgresql://pet@bmdb1.debian.org:5435/pet')
    else:
        return sqlalchemy.create_engine('postgresql://pet@bmdb1.debian.org:5435/pet?sslmode=verify-full&sslrootcert=/etc/ssl/debian/certs/ca.crt')



# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011, Ansgar Burchardt <ansgar@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from pyramid.config import Configurator
from pyramid.events import subscriber, NewRequest

from pet.models import Session
from pet.web.views import *

@subscriber(NewRequest)
def add_session_to_request(event):
  event.request.session = Session()
  def callback(request):
    request.session.rollback()
  event.request.add_finished_callback(callback)

def main(global_config=None, **settings):
  #settings.update({
  #})

  config = Configurator(settings=settings)
  config.include('pyramid_chameleon')
  config.add_static_view('static', 'pet.web:static')

  config.add_route('overview', '/{team_name}/pet.cgi')
  config.add_route('changelog', '/changelog/{named_tree_id}')
  config.add_route('notify', '/{team_name}/pet-notify.cgi')
  config.scan()

  return config.make_wsgi_app()



# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011, Ansgar Burchardt <ansgar@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from pet.classifier import Classifier
from pet.models import *

import debian.changelog

from pyramid.view import view_config
from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotFound, HTTPBadRequest
from pyramid.url import route_url
from sqlalchemy.orm import exc

import re
import os

@view_config(route_name='overview', renderer='pet.web:templates/overview.pt')
class Overview(object):
  def __init__(self, request):
    self.request = request
    self.session = request.session
    self.team_name = request.matchdict['team_name']
  def changelog_url(self, named_tree):
    return route_url('changelog', self.request, named_tree_id=named_tree.id)
  def __call__(self):
    named_trees = self.session.query(NamedTree) \
        .join(NamedTree.package).join(Package.repository) \
        .join(Repository.team).filter(Team.name == self.team_name) \
        .filter((NamedTree.type == 'branch') & (NamedTree.name == None))
    suite_cond = "1=1"
    bt_cond = "1=1"

    classifier = Classifier(self.session, named_trees, suite_cond, bt_cond)

    return {
      "classified": classifier.classify(),
      "classes": classifier.classes()
    }

@view_config(route_name='changelog', renderer='pet.web:templates/changelog.pt')
def changelog(request):
  try:
    named_tree = request.session.query(NamedTree).filter_by(id=request.matchdict['named_tree_id']).one()
    changelog_contents = named_tree.file("debian/changelog").contents
    if changelog_contents is None:
      raise HTTPNotFound()
  except exc.NoResultFound:
    raise HTTPNotFound()

  changelog = debian.changelog.Changelog(changelog_contents, max_blocks=1, strict=False)
  return { "changelog": unicode(changelog).strip() }

@view_config(route_name='notify')#, request_method='POST')
def notify(request):
  path = request.session.query(Config.value).filter_by(key='request_directory').scalar() or '/srv/pet.debian.net/requests'
  repo_name = request.params['repository']
  repo_id = request.session.query(Repository.id).join(Repository.team) \
      .filter(Repository.name == request.params['repository']) \
      .filter(Team.name == request.matchdict['team_name']).scalar()
  with open(os.path.join(path, 'update-{0}'.format(repo_id)), 'w') as fh:
    print >>fh, 'requested-by: {0}'.format(request.remote_addr)
  return Response('Ok.')



# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011, Ansgar Burchardt <ansgar@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from pet.exceptions import *
from pet.models import *
import pet.vcs
import pet.bts
import pet.watch

import debian
import debian.changelog
import debian.deb822
import os.path
import re
import shutil
import sqlalchemy.orm
import sqlalchemy.orm.exc
import subprocess
import tempfile

re_ignore = re.compile("IGNORE[ -]VERSION:?\s*(?P<version>\S+)", re.IGNORECASE)
re_waits_for = re.compile(r"""
  WAITS[ -]FOR:?\s*
  (?P<package>\S*)                           # package name
  (?:
    \s+(:?\([<=>]*\s*)?(?P<version>\S+?)\)?  # optional version number
    (?:\s+(?P<comment>.*))?                  # and comment
  )?$""", re.IGNORECASE | re.VERBOSE)
re_todo = re.compile(r"\A\s*(?:\* )?(?:TODO|PROBLEM|QUESTION):")

class NamedTreeUpdater(object):
  """update a `pet.models.NamedTree`"""
  def delete_old_files(self):
    """remove all outdated versions of files for this named tree"""
    self.session.query(File).filter((File.named_tree == self.named_tree) & (File.commit_id != self.named_tree.commit_id)).delete()
  def _get(self, filename):
    """
    get contents of a file for the current named tree as a string,
    or None if the file does not exist
    """
    if self.named_tree.type == 'tag':
      contents = self.vcs.file(self.package.name, filename, tag=self.named_tree.name)
    elif self.named_tree.type == 'branch':
      contents = self.vcs.file(self.package.name, filename, branch=self.named_tree.name)
    else:
      raise ValueError("unknown NamedTree type '{0}'".format(self.named_tree.type))

    if contents is not None:
      try:
        contents = unicode(contents, 'utf-8')
      except UnicodeDecodeError:
        contents = unicode(contents, 'iso-8859-1')
    return contents
  def file(self, filename):
    """retrieve the named file from the VCS and store it in the database.

    Returns a tuple (file, changed) with file a `pet.models.File`
    object and changed a Boolean indicating that the file was
    updated.
    """
    try:
      f = self.named_tree.file(filename)
      changed = False
    except sqlalchemy.orm.exc.NoResultFound:
      try:
        contents = self._get(filename)
      except FileNotFound:
        contents = None
      f = File(named_tree=self.named_tree, commit_id=self.named_tree.commit_id, name=filename, contents=contents)
      self.session.add(f)
      changed = True
    return f, changed
  def update_patches(self):
    """update list of patches for named tree"""
    patches, changed = self.file("debian/patches/series")
    if not changed and not self.force: return
    self.session.query(Patch).filter_by(named_tree=self.named_tree).delete()
    if patches.contents:
      for line in patches.contents.splitlines():
        if line.startswith("#"):
          continue
        fields = line.split()
        if len(fields):
          patch = Patch(named_tree=self.named_tree, name=fields[0])
          self.session.add(patch)
  def update_control(self):
    """update information extracted from debian/control"""
    control_file, changed = self.file("debian/control")
    if not changed and not self.force: return
    nt = self.named_tree
    if control_file.contents:
      control = debian.deb822.Deb822(control_file.contents)
      nt.source = control.get("Source")
      if "Maintainer" in control:
        nt.maintainer = control["Maintainer"].strip()
      else:
        nt.maintainer = None
      if "Uploaders" in control:
        nt.uploaders = [ u.strip() for u in control["Uploaders"].split(",") ]
      else:
        nt.uploaders = None
      if "Homepage" in control:
        nt.homepage = control["Homepage"].strip()
      else:
        nt.homepage = None
    else:
      nt.source = nt.maintainer = nt.uploaders = nt.homepage = None
  def update_changelog(self):
    """update information extracted from debian/changelog"""
    changelog_file, changed = self.file("debian/changelog")
    if not changed and not self.force: return
    nt = self.named_tree
    nt.ignore = nt.todo = False
    self.session.query(Wait).filter_by(named_tree=self.named_tree).delete()

    if changelog_file.contents:
      changelog = debian.changelog.Changelog(changelog_file.contents,
          strict=False)
      nt.source_changelog = changelog.package
      nt.version = str(changelog.version)
      nt.versions = [ str(v) for v in changelog.versions ]
      nt.distribution = changelog.distributions
      nt.urgency = changelog.urgency
      nt.last_changed = changelog.date
      nt.last_changed_by = changelog.author

      # TODO: Use public API once #634849 is fixed.
      for line in changelog._blocks[0].changes():
        match = re_ignore.search(line)
        if match and match.group('version') == nt.version:
          nt.ignore = True

        match = re_waits_for.search(line)
        if match:
          wait = Wait(named_tree=self.named_tree, name=match.group('package'), version=match.group('version'), comment=match.group('comment'))
          self.session.add(wait)

        if re_todo.search(line):
          nt.todo = True
    else:
      nt.source_changelog = nt.version = nt.distribution = nt.urgency = nt.last_changed = nt.last_changed_by = None
      nt.versions = []
  def update_watch(self):
    """update cached version of debian/watch"""
    watch_file, changed = self.file("debian/watch")
  def run(self, named_tree, package, vcs, force=False):
    self.session = Session.object_session(named_tree)
    self.named_tree = named_tree
    self.package = package
    self.vcs = vcs
    self.force = force

    print "I: updating {0}, {1} {2}".format(self.package.name, self.named_tree.type, self.named_tree.name)

    self.delete_old_files()
    self.update_patches()
    self.update_control()
    self.update_changelog()
    self.update_watch()

class PackageUpdater(object):
  """update a `pet.models.Package`"""
  def _update_named_tree_list(self, type, known, existing):
    changed = []

    for name, nt in known.iteritems():
      commit_id = existing.get(name, None)
      if commit_id is None:
        self.session.delete(nt)
      else:
        nt.commit_id = str(commit_id)
        changed.append(nt)

    for name, commit_id in existing.iteritems():
      if name not in known:
        nt = NamedTree(type=type, name=name, commit_id=str(commit_id), package=self.package)
        self.session.add(nt)
        changed.append(nt)

    return changed
  def update_tag_list(self):
    known = self.package.tags
    existing = self.vcs.tags(self.package.name)
    return self._update_named_tree_list('tag', known, existing)
  def update_branch_list(self):
    known = self.package.branches
    existing = self.vcs.branches(self.package.name)
    return self._update_named_tree_list('branch', known, existing)
  def update_named_trees(self):
    ntu = NamedTreeUpdater()
    for nt in self.package.named_trees:
      ntu.run(nt, self.package, self.vcs, force=self.force)
  def run(self, package, vcs, force=False):
    self.session = Session.object_session(package)
    self.package = package
    self.vcs = vcs
    self.force = force

    self.update_tag_list()
    self.update_branch_list()
    self.update_named_trees()

class RepositoryUpdater(object):
  """update a `pet.models.Repository`"""
  def __init__(self, repository, force=False):
    self.session = Session.object_session(repository)
    self.repository = repository
    self.vcs = pet.vcs.vcs_backend(repository)
    self.force = force
  def update_package_list(self):
    self.session.begin_nested()
    try:
      known_packages = {}
      for p in self.repository.packages:
        known_packages[p.name] = p
      existing_packages = self.vcs.packages

      for name, p in known_packages.iteritems():
        if name not in existing_packages:
          self.session.delete(p)
      for name in existing_packages:
        if name not in known_packages:
          package = Package(name=name, repository=self.repository)
          self.session.add(package)
      self.session.commit()
    except:
      self.session.rollback()
      raise
  def update_changed_packages(self):
    self.session.begin_nested()
    try:
      named_trees_by_package = {}
      for p in self.repository.packages:
        named_trees_by_package[p] = []
      for nt in self.session.query(NamedTree).join(NamedTree.package).filter(Package.repository==self.repository):
        named_trees_by_package[nt.package].append(nt)
      print "D: Looking for changes in {0} packages.".format(len(named_trees_by_package))
      changed = self.vcs.changed_named_trees(self.session, named_trees_by_package)
      print "D: Found {0} changed packages.".format(len(changed))

      ntu = NamedTreeUpdater()
      for p, nts in changed.iteritems():
        for nt in nts:
          ntu.run(nt, p, self.vcs)
      self.session.commit()
    except:
      self.session.rollback()
      raise
  def update_all_packages(self):
    for p in self.repository.packages:
      pu = PackageUpdater()
      self.session.begin_nested()
      try:
        print "I: Updating package {0}".format(p.name)
        pu.run(p, self.vcs, force=self.force)
        self.session.commit()
      # XXX: Do we want to catch all exceptions here? Probably yes.
      #except Exception as e:
      #  self.session.rollback()
      #  print "E: error while updating package {0}: {1}".format(p.name, e)
      except:
        self.session.rollback()
        raise
  def run(self):
    self.update_package_list()
    if self.force:
      self.update_all_packages()
    else:
      self.update_changed_packages()

class SuiteUpdater(object):
  def __init__(self, suite, archive, tmpdir=None):
    self.session = Session.object_session(suite)
    self.suite = suite
    self.archive = archive
    if tmpdir:
      self.tmpdir = tmpdir
      self.cleantmp = False
    else:
      self.tmpdir = tempfile.mkdtemp(prefix="pet-suite-{0}".format(self.suite.name))
      self.cleantmp = True
  def __del__(self):
    if self.cleantmp:
      shutil.rmtree(self.tmpdir)
  def delete_package_list(self):
    self.session.query(SuitePackage).filter_by(suite=self.suite).delete()
  def _download(self, source, target):
    target_xz = target + ".xz"
    r = subprocess.call(['wget', '--quiet', '-O', target_xz, '--', source])
    if r:
      raise IOError("wget failed for {0}.".format(source))
    with open(target, 'w') as fh:
      r = subprocess.call(['xz', '--decompress', '--stdout', '--', target_xz], stdout=fh)
      if r:
        raise IOError("xz failed for {0}.".format(source))
  def add_package_list(self):
    for component in self.suite.components:
      url = "{0}/dists/{1}/{2}/source/Sources.xz".format(self.archive.url, self.suite.name, component)
      target = os.path.join(self.tmpdir, "sources-{0}-{1}-{2}".format(self.archive.id, self.suite.id, component))
      self._download(url, target)
      with open(target, 'r') as fh:
        for s in debian.deb822.Sources.iter_paragraphs(fh):
          if "Uploaders" in s:
            uploaders = [ u.strip() for u in s["Uploaders"].split(",") ]
          else:
            uploaders = None
          sp = SuitePackage(
            suite      = self.suite,
            source     = s["Package"],
            version    = s["Version"],
            component  = component,
            maintainer = s["Maintainer"].strip(),
            uploaders  = uploaders,
            dsc        = "{0}/{1}_{2}.dsc".format(s["Directory"], s["Package"], s["Version"]),
            )
          self.session.add(sp)
  def run(self):
    self.session.begin_nested()
    try:
      self.delete_package_list()
      self.add_package_list()
      self.session.commit()
    except:
      self.session.rollback()
      raise

class ArchiveUpdater(object):
  def __init__(self, archive):
    self.archive = archive
  def run(self):
    for suite in self.archive.suites:
      su = SuiteUpdater(suite, self.archive)
      su.run()

class BugTrackerUpdater(object):
  def __init__(self, bug_tracker):
    self.session = Session.object_session(bug_tracker)
    self.bug_tracker = bug_tracker
  def _delete_unreferenced_bugs(self, sources):
    # TODO: Should use SQL to look for source package names in named_trees.
    print self.session.query(Bug).join(Bug.bug_sources) \
        .filter((Bug.bug_tracker==self.bug_tracker) & ~ BugSource.source.in_(sources)) \
        .statement
  def _update_bugs(self, bug_reports):
    bugs = {}
    for bug in self.session.query(Bug).filter_by(bug_tracker=self.bug_tracker):
      bugs[bug.bug_number] = bug

    print "I: Updating {0} bug reports...".format(len(bug_reports))
    progress = 0
    for br in bug_reports:
      # TODO: one query per bug is SLOOOOOW!
      try:
        bug = bugs[br.bug_number]
      except KeyError:
        bug = Bug(bug_tracker=self.bug_tracker, bug_number=br.bug_number)
        self.session.add(bug)
      br.update_bug(bug)

      progress += 1
      if progress % 10 == 0:
        print "D:   {0} / {1} done".format(progress, len(bug_reports))

  def run(self, named_trees=None):
    # TODO: Add binary_source_map
    bts = pet.bts.DebianBugTracker({}, ignore_unknown_binaries=True)
    # TODO: Unify code path once _delete_unreferenced_bugs is fixed
    # to no longer need the list of sources.
    if named_trees is None:
      sources = [ s[0] for s in self.session.query(NamedTree.source).distinct() ]
      self._delete_unreferenced_bugs(sources)
      bug_numbers = [ b[0] for b in self.session.query(Bug.bug_number).filter_by(bug_tracker=self.bug_tracker).all() ]
      bug_reports = bts.search(sources, bug_numbers)
      self._update_bugs(bug_reports)
    else:
      sources = list(set([ nt.source for nt in named_trees ]))
      bug_reports = bts.search(sources)
      self._update_bugs(bug_reports)

class WatchUpdater(object):
  def __init__(self, session):
    self.session = session
    self.watcher = pet.watch.Watcher()

  def update_watch(self, watch):
    if watch.contents is None:
      return
    result = self.watcher.check(watch.contents)
    if result['errors'] is None:
      wr = WatchResult(named_tree=watch.named_tree, homepage=result['homepage'], upstream_version=str(result['version']), download_url=result['url'], debian_version=result['dversionmangle'](watch.named_tree.version))
    else:
      error = ", ".join([ str(e) for e in result['errors'] ])
      wr = WatchResult(named_tree=watch.named_tree, homepage=result.get('homepage'), error=error)
    self.session.add(wr)
  def run(self, named_trees=None):
    self.session.begin_nested()
    try:
      if named_trees is None:
        named_trees = self.session.query(NamedTree)
      named_trees = named_trees.filter((NamedTree.type=='branch') & (NamedTree.name == None))
      watches = self.session.query(File) \
          .filter(File.name == 'debian/watch') \
          .filter(File.named_tree_id.in_(named_trees.from_self(NamedTree.id).subquery())) \
          .options(sqlalchemy.orm.joinedload(File.named_tree))
      self.session.query(WatchResult) \
          .filter(WatchResult.named_tree_id.in_(named_trees.from_self(NamedTree.id).subquery())).delete(False)
      for watch in watches:
        print "D: checking watch for {0}".format(watch.named_tree.source)
        self.update_watch(watch)
    except:
      self.session.rollback()
      raise
    self.session.commit()

class Updater(object):
  def run(self):
    session = Session()
    try:
      for r in session.query(Repository).all():
        ru = RepositoryUpdater(r)
        ru.run()
      session.commit()
    except:
      session.rollback()
      raise



# vim:ts=2:sw=2:et:ai:sts=2
# Copyright 2011, Ansgar Burchardt <ansgar@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import pet
import sqlalchemy.dialects.postgresql
import sqlalchemy.ext.declarative
import sqlalchemy.orm
import sqlalchemy.schema
import sqlalchemy.types
import os.path

def engine(no_cert):
    if(no_cert):
        return sqlalchemy.create_engine('postgresql://pet@bmdb1.debian.org:5435/pet')
    else:
        return sqlalchemy.create_engine('postgresql://pet@bmdb1.debian.org:5435/pet?sslmode=verify-full&sslrootcert=/etc/ssl/debian/certs/ca.crt')

class DebVersion(sqlalchemy.types.UserDefinedType):
  """database type for PostgreSQL's debversion type"""
  def get_col_spec(self):
    return "DEBVERSION"
  def bind_processor(self, dialect):
    def process(value):
      return value
    return process
  def result_processor(self, dialect, coltype):
    def process(value):
      return value
    return process

# XXX: Shouldn't there be an API for this?
sqlalchemy.dialects.postgresql.base.ischema_names['debversion'] = DebVersion

# check if certificate file exists, if it exists, the no-certificate flag is
# false (we use the certificate). If if does not exist, the no-certificate
# flag is true (we dont use the certificat).
if(os.path.isfile("/etc/ssl/debian/certs/ca.crt")):
    engine = engine(False)
else:
    print('No certification file found (/etc/ssl/debian/certs/ca.crt)')
    continue_without_ssl = raw_input('Do you want do continue? [Y/n] ')
    if (continue_without_ssl is not 'y'):
        engine = engine(False)
    else:
        engine = engine(True)
metadata = sqlalchemy.schema.MetaData()
metadata.reflect(bind=engine)
Session = sqlalchemy.orm.sessionmaker(bind=engine)
Base = sqlalchemy.ext.declarative.declarative_base(metadata=metadata)

class Config(Base):
  """configuration settings"""
  __tablename__ = 'config'

class Team(Base):
  """model for a team

  A team owns one more more `Repository`s.
  """
  __tablename__ = 'team'

class Repository(Base):
  """model for a repository

  A `Repository` represents a group of packages stored in a common VCS
  location, such as a Subversion repository or a directory of Git
  repositories.
  """
  __tablename__ = 'repository'
  team = sqlalchemy.orm.relation('Team', backref='repositories')
  @property
  def vcs(self):
    """returns the version control backend for the repository.

    See `pet.vcs` for more information about the returned object.
    """
    if '_vcs' not in self.__dict__:
      from pet.vcs import vcs_backend
      self._vcs = vcs_backend(self)
    return self._vcs

class Package(Base):
  """model for a package

  A `Package` represents a single (source) package in a `Repository`.
  Here "package" refers to the directory struckture in the VCS, the
  name can differ from the Debian package name.
  """
  __tablename__ = 'package'
  repository = sqlalchemy.orm.relation('Repository', backref='packages')
  @property
  def trunk(self):
    return self.branches[None]
  def _named_trees(self, type):
    session = Session.object_session(self)
    named_trees = {}
    for nt in session.query(NamedTree).filter_by(package=self, type=type):
      named_trees[nt.name] = nt
    return named_trees
  @property
  def branches(self):
    """returns a list of branches as `NamedTree` objects"""
    return self._named_trees('branch')
  @property
  def tags(self):
    """returns a list of tags as `NamedTree` objects"""
    return self._named_trees('tag')

class NamedTree(Base):
  """model for branches and tags

  This class represents branches and tags.  The trunk is available as
  the branch with name None.
  """
  __tablename__ = 'named_tree'
  package = sqlalchemy.orm.relation('Package', lazy='joined',
      backref=sqlalchemy.orm.backref('named_trees',  passive_deletes=True))
  def _file(self, filename):
    session = Session.object_session(self)
    return session.query(File).filter_by(named_tree=self, name=filename, commit_id=self.commit_id)
  def has_file(self, filename):
    return self._file(filename).count() != 0
  def file(self, filename):
    """retrieve a file from the cache"""
    return self._file(filename).one()
  def link(self, filename, directory=False):
    """provide a link to the VCS web interface"""
    return self.package.repository.vcs.link(self.package.name, filename, directory, named_tree=self)

class WatchResult(Base):
  """result of checking debian/watch

  This class represents the results of checking debian/watch for
  upstream versions.  It provides information about the latest
  upstream version available.
  """
  __tablename__ = 'watch_result'
  named_tree = sqlalchemy.orm.relation('NamedTree',
      backref=sqlalchemy.orm.backref('watch_result', uselist=False))

class Wait(Base):
  """model for WAITS-FOR entries

  This class is used to keep track of WAITS-FOR entries in
  debian/changelog.  It gives the package/version we are waiting for.
  """
  __tablename__ = 'wait'
  named_tree = sqlalchemy.orm.relation('NamedTree', backref=sqlalchemy.orm.backref('waits', passive_deletes=True))

class File(Base):
  """file retrieved from a VCS

  This class represents a file belonging to a `NamedTree`.
  """
  __tablename__ = 'file'
  named_tree = sqlalchemy.orm.relation('NamedTree',
      backref=sqlalchemy.orm.backref('files', passive_deletes=True))

class Patch(Base):
  """patches present in debian/patches/series"""
  __tablename__ = 'patch'
  named_tree = sqlalchemy.orm.relation('NamedTree',
      backref=sqlalchemy.orm.backref('patches', passive_deletes=True))

class Archive(Base):
  """model for an archive"""
  __tablename__ = 'archive'

class Suite(Base):
  """model for a suite

  This class represents a suite in an `Archive`.
  """
  __tablename__ = 'suite'
  archive = sqlalchemy.orm.relation('Archive', backref='suites')

class SuitePackage(Base):
  """model for a source package in a suite

  This class represents a source package in a `Suite`.
  """
  __tablename__ = 'suite_package'
  suite = sqlalchemy.orm.relation('Suite', backref='packages')

class SuiteBinary(Base):
  """model for a binary package in a suite

  This class represents a binary package in a `Suite`.
  """
  __tablename__ = 'suite_binary'
  source = sqlalchemy.orm.relation('SuitePackage', backref='binaries')
  """returns the source package for this binary"""

class BugTracker(Base):
  """model for a bug tracker

  This class represents a bug tracker such as Debian's BTS or Ubuntu's
  bug tracker component in Launchpad.
  """
  __tablename__ = 'bug_tracker'

class Bug(Base):
  """model for a bug

  A `Bug` represents a single bug report retrieved from a
  `BugTracker`.
  """
  __tablename__ = 'bug'
  bug_tracker = sqlalchemy.orm.relation('BugTracker', backref='bugs')

class BugSource(Base):
  """associates bugs with source packages

  This class is used to associate `Bug` objects with source package
  names.  It also provides information about affected versions.
  """
  __tablename__ = 'bug_source'
  bug = sqlalchemy.orm.relation('Bug', backref='bug_sources')
